###############################################################################
##                                                                           ##
## This file is part of ModelBlocks. Copyright 2009, ModelBlocks developers. ##
##                                                                           ##
##    ModelBlocks is free software: you can redistribute it and/or modify    ##
##    it under the terms of the GNU General Public License as published by   ##
##    the Free Software Foundation, either version 3 of the License, or      ##
##    (at your option) any later version.                                    ##
##                                                                           ##
##    ModelBlocks is distributed in the hope that it will be useful,         ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of         ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          ##
##    GNU General Public License for more details.                           ##
##                                                                           ##
##    You should have received a copy of the GNU General Public License      ##
##    along with ModelBlocks.  If not, see <http://www.gnu.org/licenses/>.   ##
##                                                                           ##
###############################################################################

#!/usr/bin/ruby

#####################################################################
# cnftrees2cedepths.rb
#  looks for and annotates center-embedding depths and Left-Right
#   constituency
#  only works on binary-branching trees
#
######################################################################

require "scripts/umnlp.rb"

class Tree
  def propDown

    if @head =~ /.*\^([LR]),([0-9]+)/
      depth = $2.to_i
      side  = $1
    else
      depth = 1
      side  = "L"
      @head += "^#{side},#{depth}"
    end

    if @children.size() <= 1
      return
    end
    if side=="R"
      ldepth = depth+1
    else
      ldepth = depth
    end
    @children[0].head += "^L,#{ldepth}"
    @children[0].propDown

    if @children.size() <= 1
      return
    end
    #depth += 1
    @children[1].head += "^R,#{depth}"
    @children[1].propDown

  end
end

while(line = gets)
  t = Tree.new(line)
  t.propDown
  if t.prob == nil
    print t.to_s + "\n"
  else
    puts "#{t.to_s} : #{t.prob}"
  end
end
